//
//  AUManager.m
//  AppTemplate
//
//  Created by Vesa Norilo on 2/20/12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#include <stdexcept>
#include <cstdio>

#import <AudioUnit/AudioUnit.h>
#import <AudioToolbox/AudioToolbox.h>
#import <AudioToolbox/AudioServices.h>
#import <UIKit/UIKit.h>
#import "AUStream.h"


#if TARGET_IPHONE_SIMULATOR
#define AU_NOMINAL_LEVEL   1
#define AU_SAMPLE   Float32
#define AU_FORMATFLAGS kAudioFormatFlagsNativeFloatPacked
#else
#define AU_NOMINAL_LEVEL (1 << kAudioUnitSampleFractionBits)
#define AU_SAMPLE   SInt32
#define AU_FORMATFLAGS kAudioFormatFlagsCanonical
#endif

#pragma mark AudioSession listeners
void interruptionListener(	void *	inClientData,
						  UInt32	inInterruptionState)
{
}

void propListener(	void *                  inClientData,
				  AudioSessionPropertyID	inID,
				  UInt32                  inDataSize,
				  const void *            inData)
{
}

void SetupDefaultAudioSession()
{
 
}


static char errorBuffer[256];
#define EXCEPTION(AUCALL, MESSAGE) {OSStatus err; if ((err=AUCALL)!=0) {snprintf(errorBuffer,256,"%s: %i",MESSAGE,(int)err);throw std::runtime_error(errorBuffer);}}

struct AudioFormat : public AudioStreamBasicDescription{
    AudioFormat(int SampleRate, int nChannels)
    {
        mFormatID = kAudioFormatLinearPCM;
        mFormatFlags = AU_FORMATFLAGS;
        int sampleSize = sizeof(AU_SAMPLE);
        mBitsPerChannel = 8 * sampleSize;
        mChannelsPerFrame = nChannels;
        mFramesPerPacket = 1;
        mBytesPerPacket = mBytesPerFrame = nChannels * sampleSize;
        mSampleRate = SampleRate;
        mReserved = 0;
    }
};

@interface AUStream()
@property (readonly) AudioCallbackFunction RenderProcedure;
@property (readonly) void* RenderData;
@property (readonly) float* ConversionBuffer;
@end

static float inline Clip(float val)
{
    if (val>1) return 1;
    else if (val<-1) return -1;
    else return val;
}

static OSStatus	AUStreamCallback(
							void						*inRefCon, 
							AudioUnitRenderActionFlags 	*ioActionFlags, 
							const AudioTimeStamp 		*inTimeStamp, 
							UInt32 						inBusNumber, 
							UInt32 						inNumberFrames,
							AudioBufferList 			*ioData)
{
	AUStream *THIS = (__bridge AUStream *)inRefCon;
	
    OSStatus err = AudioUnitRender(THIS.RemoteIO, ioActionFlags, inTimeStamp, 1, inNumberFrames, ioData);
	if (err) { printf("PerformThru: error %d\n", (int)err); return err; }
		
    AU_SAMPLE *data_ptr = (AU_SAMPLE *)(ioData->mBuffers[0].mData);
    
    int todo = inNumberFrames;
    
    float *buffer = THIS.ConversionBuffer;
    int numChannels = THIS.numChannels;
    void *renderData = THIS.RenderData;
    int bufferLimit = CONVERT_BUFFER_SIZE / numChannels;
    
    while(todo>0)
    {
        int now = todo;
        if (now > bufferLimit) now = bufferLimit;
        
        for(int i=0;i<now*numChannels;i++) buffer[i] = data_ptr[i] / (float)AU_NOMINAL_LEVEL;
        THIS.RenderProcedure(buffer,numChannels,now,renderData);
        for(int i=0;i<now*numChannels;i++) data_ptr[i] = (AU_SAMPLE)(Clip(buffer[i]) * (float)AU_NOMINAL_LEVEL);
        
        data_ptr += now * numChannels;
        todo-=now;
    }
	return err;
}

@implementation AUStream
@synthesize RemoteIO = _RemoteIO;
@synthesize started = _started;
@synthesize numChannels = _numChannels;
@synthesize sampleRate = _sampleRate;
@synthesize RenderData = _renderData;
@synthesize RenderProcedure = _renderProc;
@dynamic ConversionBuffer;

-(float*)ConversionBuffer{
    return conversionBuffer;
}

-(id)initWithCallback:(AudioCallbackFunction)RenderProc RefCon:(void*)renderData SampleRate:(int)sr Channels:(int)chans
{
    AudioFormat StreamFormat(sr,chans);
    
	try {		
        EXCEPTION(AudioSessionInitialize(NULL, NULL, interruptionListener, (__bridge void*)self),
                  "Can't initialize audio session!");
            UInt32 category = kAudioSessionCategory_PlayAndRecord;
            EXCEPTION(AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(category), &category),
                      "Couldn't set audio session category");
            
            EXCEPTION(AudioSessionAddPropertyListener(kAudioSessionProperty_AudioRouteChange, propListener, (__bridge void*)self),
                      "Couldn't add an audio session property listener");

            UInt32 inputAvailable = 0;
            UInt32 size = sizeof(inputAvailable);
            
            // we do not want to allow recording if input is not available
            EXCEPTION(AudioSessionGetProperty(kAudioSessionProperty_AudioInputAvailable, &size, &inputAvailable),
                      "Couldn't check for audio input availability");
            
            // we also need to listen to see if input availability changes
            EXCEPTION(AudioSessionAddPropertyListener(kAudioSessionProperty_AudioInputAvailable, propListener, (__bridge void*)self),
                      "Couldn't set up an audio input availability listener");
            
            // route audio to speaker
            UInt32 doChangeDefaultRoute = 1;
            AudioSessionSetProperty (
                                     kAudioSessionProperty_OverrideCategoryDefaultToSpeaker,
                                     sizeof (doChangeDefaultRoute),
                                     &doChangeDefaultRoute);
            
            // allow bluetooth input
            UInt32 allowBluetoothInput = 1;
            
            AudioSessionSetProperty (
                                     kAudioSessionProperty_OverrideCategoryEnableBluetoothInput,
                                     sizeof (allowBluetoothInput),
                                     &allowBluetoothInput);
            
            EXCEPTION(AudioSessionSetActive(true),"Couldn't activate audio session");
		// Open the output unit
		AudioComponentDescription desc;
		desc.componentType = kAudioUnitType_Output;
		desc.componentSubType = kAudioUnitSubType_RemoteIO;
		desc.componentManufacturer = kAudioUnitManufacturer_Apple;
		desc.componentFlags = 0;
		desc.componentFlagsMask = 0;
		
		AudioComponent comp = AudioComponentFindNext(NULL, &desc);
		
		EXCEPTION(AudioComponentInstanceNew(comp, &_RemoteIO), "couldn't open the remote I/O unit");
        
		UInt32 one = 1;
		EXCEPTION(AudioUnitSetProperty(_RemoteIO, kAudioOutputUnitProperty_EnableIO, kAudioUnitScope_Input, 1, &one, sizeof(one)), "couldn't enable input on the remote I/O unit");

        CallbackStruct.inputProc = AUStreamCallback;
        CallbackStruct.inputProcRefCon = (__bridge void*)self;
        
		EXCEPTION(AudioUnitSetProperty(_RemoteIO, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, 0, &CallbackStruct, sizeof(CallbackStruct)), "couldn't set remote i/o render callback");
		
        // set our required format - Canonical AU format: LPCM non-interleaved 8.24 fixed point
		EXCEPTION(AudioUnitSetProperty(_RemoteIO, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &StreamFormat, sizeof(StreamFormat)), "couldn't set the remote I/O unit's output client format");
		EXCEPTION(AudioUnitSetProperty(_RemoteIO, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 1, &StreamFormat, sizeof(StreamFormat)), "couldn't set the remote I/O unit's input client format");
        
		EXCEPTION(AudioUnitInitialize(_RemoteIO), "couldn't initialize the remote I/O unit");
        
        _started = NO;
        _numChannels = chans;
        _sampleRate = sr;
        _renderData = renderData;
        _renderProc = RenderProc;
	}
	catch (std::exception &e) {
		fprintf(stderr, "Opening remote I/O: %s \n", e.what());
		return nil;
	}
	catch (...) {
		fprintf(stderr, "Opening remote I/O: An unknown error occurred\n");
		return nil;
	}		
	return [super init];
}

+(id)streamWithCallback:(AudioCallbackFunction)render  RefCon:(void*)renderData SampleRate:(int)sr Channels:(int)chans
{
#if __has_feature(objc_arc)
    return [[AUStream alloc] initWithCallback:render RefCon:renderData SampleRate:sr Channels:chans];
#else
    return [[[AUStream alloc] initWithCallback:render RefCon:renderData SampleRate:sr Channels:chans] autorelease];
#endif
}

-(void)dealloc{
    [self stop];
    try{
        EXCEPTION(AudioUnitUninitialize(_RemoteIO),"couldn't uninitialize the remote I/O unit");
    }
	catch (std::exception &e) {
		fprintf(stderr, "Closing remote I/O: %s \n", e.what());
	}
	catch (...) {
		fprintf(stderr, "Closing remote I/O: An unknown error occurred\n");
	}		
}

-(BOOL)play{
    if (_started) return NO;
    try{
        EXCEPTION(AudioOutputUnitStart(_RemoteIO), "couldn't start remote I/O streaming");        
        _started = YES;
        return YES;
    }
	catch (std::exception &e) {
		fprintf(stderr, "Starting remote I/O: %s \n", e.what());
	}
	catch (...) {
		fprintf(stderr, "Starting remote I/O: An unknown error occurred\n");
	}		
    return NO;
}

-(BOOL)stop{
    if (!_started) return NO;
    try{
        EXCEPTION(AudioOutputUnitStop(_RemoteIO), "couldn't stop remote I/O streaming");        
        _started = NO;
        return YES;
    }
	catch (std::exception &e) {
		fprintf(stderr, "Stopping remote I/O: %s \n", e.what());
	}
	catch (...) {
		fprintf(stderr, "Stopping remote I/O: An unknown error occurred\n");
	}		
    return NO;
}

@end
